'''pyKML Parser Module

The pykml.parser module provides functions that can be used to parse KML 
from a file or remote URL.
'''
import os
from lxml import etree, objectify

class Schema():
    "A class representing an XML Schema used to validate KML documents"
    def __init__(self, schema):
        module_dir = os.path.split(__file__)[0]   # get the module path
        schema_file = os.path.join(module_dir, "schemas", schema)
        with open(schema_file) as f:
            self.schema = etree.XMLSchema(file=f)
    
    def validate(self, doc):
        """Validates a KML document
        
        This method eturns a boolean value indicating whether the KML document 
        is valid when compared to the XML Schema."""
        return self.schema.validate(doc)
    
    def assertValid(self, doc):
        """Asserts that a KML document is valide
        
        The method generates a validation error if the document is not valid
        when compared to the XML Schema.
        """
        return self.schema.assertValid(doc)

def fromstring(text, schema=None):
    """Parses a KML text string
    
    This function parses a KML text string and optionally validates it against 
    a provided schema object"""
    if schema:
        parser = objectify.makeparser(schema = schema.schema)
        return objectify.fromstring(text, parser=parser)
    else:
        return objectify.fromstring(text)

def parse(fileobject, schema=None):
    """Parses a file object
    
    This functon parses a KML file object, and optionally validates it against 
    a provided schema.
    """
    if schema:
        # with validation
        parser = objectify.makeparser(schema = schema.schema, strip_cdata=False)
        return objectify.parse(fileobject, parser=parser)
    else:
        # without validation
        return objectify.parse(fileobject)



"""
Conversions between WGS84, Cartesian, Mercator and TMS coordinates and bounding boxes,
useful for converting geographic view ports as generated by Google Earth to TMS bounding boxes
"""

import math
import itertools
from abc import ABCMeta, abstractmethod

__author__ = "Martin Loetzsch, Gregor Sturm"
__licence__ = "Apache 2.0"

_earthradius = 6378137.0  # m

_tilesize = _initialresolution = _originshift = None
_originshift_180 = _180_originshift = None

_pi_2 = math.pi / 2.0
_pi_180 = math.pi / 180.0
_pi_360 = math.pi / 360.0
_180_pi = 180.0 / math.pi


def init_geometry(tilesize=256.0):
    global _tilesize, _initialresolution, _originshift
    global _originshift_180, _180_originshift
    _tilesize = tilesize
    _initialresolution = 2 * math.pi * _earthradius / _tilesize
    _originshift = 2 * math.pi * _earthradius / 2.0
    _originshift_180 = _originshift / 180.0
    _180_originshift = 180.0 / _originshift


init_geometry()


def griditer(x, y, ncol, nrow=None, step=1):
    """
    Iterate through a grid of tiles.

    Args:
        x (int): x start-coordinate
        y (int): y start-coordinate
        ncol (int): number of tile columns
        nrow (int): number of tile rows. If not specified, this
            defaults to ncol, s.t. a quadratic region is
            generated
        step (int): clear. Analogous to range().

    Yields:
        Tuple: all tuples (x, y) in the region delimited by
            (x, y), (x + ncol, y + ncol).

    """
    if nrow is None:
        nrow = ncol
    yield from itertools.product(range(x, x + ncol, step),
                                 range(y, y + nrow, step))


def bboxiter(tile_bounds, tiles_per_row_per_region=1):
    """
    Iterate through a grid of regions defined by a TileBB.

    Args:
        tile_bounds (GridBB):
        tiles_per_row_per_region: Combine multiple tiles in one region.
            E.g. if set to two, four tiles will be combined in one region.
            See `kml` module description for more details. Leaving the
            default value '1' simply yields all tiles in the bounding box.

    Note:
        If the number of regions would not be an integer due to specification
        of the `tiles_per_row_per_region` parameter, the boundaries will
        be rounded to the next smaller or next larger integer respectively.

        Example:
        We have the following bounding box with size 2x2 and set
        tiles_per_row_per_region = 2, delimited by the coordinates (x, y)::

             (5,5)--- ---
                 |       |
                 |       |
                 |       |
                  --- ---(9,7)

        Although this could be represented in one single region with two
        tiles per row, it will create four regions::

             (2,2)--- ---         (5/2 = 2.5 -> 2, 5/2 = 2.5 -> 2)
                 |   |   |
                  --- ---
                 |   |   |
                  --- ---(5,4)    (9/2 = 4.5 -> 5, 7/2 = 3.5 -> 4)


    Yields:
        Tuple: all tuples (x, y) in the region delimited by the TileBB

    """
    x_lower = math.floor(tile_bounds.min.x / tiles_per_row_per_region)
    y_lower = math.floor(tile_bounds.min.y / tiles_per_row_per_region)
    ncol = math.ceil(tile_bounds.max.x / tiles_per_row_per_region) - x_lower
    nrow = math.ceil(tile_bounds.max.y / tiles_per_row_per_region) - y_lower
    yield from griditer(x_lower, y_lower, ncol, nrow)


class GeographicCoordinate:
    """
    Represents a WGS84 Datum

    Args:
        lon: longitude in degrees
        lat: latitude in degrees
        height: height in meters above the surface of the earth spheroid
    """

    def __init__(self, lon=None, lat=None, height=0.0):
        self.lon = lon
        self.lat = lat
        self.height = height

    def to_cartesian(self):
        r = (_earthradius + self.height)
        cosx = math.cos(_pi_180 * self.lon)
        cosy = math.cos(_pi_180 * self.lat)
        sinx = math.sin(_pi_180 * self.lon)
        siny = math.sin(_pi_180 * self.lat)
        return CartesianCoordinate(r * cosy * cosx, r * cosy * sinx, r * siny)

    def to_mercator(self):
        return MercatorCoordinate(
            self.lon * _originshift_180,
            math.log(math.tan((90 + self.lat) * _pi_360)) / _pi_180 * _originshift_180)

    def __str__(self):
        return "<lon: " + str(self.lon) + ", lat: " + str(self.lat) + ", height: " \
               + str(self.height) + ">"


class GeographicBB:
    """ A bounding box defined by two geographic coordinates """

    def __init__(self, min_lon=None, min_lat=None, max_lon=None, max_lat=None):
        self.min = GeographicCoordinate(min_lon, min_lat)
        self.max = GeographicCoordinate(max_lon, max_lat)

    def intersects(self, other):
        return self.min.lon <= other.max.lon and self.max.lon >= other.min.lon and \
                     self.min.lat <= other.max.lat and self.max.lat >= other.min.lat

    def intersection(self, other):
        intersects = self.intersects(other)
        if intersects:
            return GeographicBB(max(self.min.lon, other.min.lon),
                                max(self.min.lat, other.min.lat),
                                min(self.max.lon, other.max.lon),
                                min(self.max.lat, other.max.lat))
        elif self.max.lon > 180.0:
            return GeographicBB(-180, self.min.lat,
                                self.max.lon - 360.0, self.max.lat).intersection(other)

    def center(self):
        return GeographicCoordinate(self.min.lon + (self.max.lon - self.min.lon) / 2,
                                    self.min.lat + (self.max.lat - self.min.lat) / 2)

    def to_mercator(self):
        return MercatorBB(self.min.to_mercator(), self.max.to_mercator())

    def __str__(self):
        return "<geographic min: " + str(self.min) + ", max: " + str(self.max) + ">"


class CartesianCoordinate:
    """ Represents a coordinate in a geocentric Cartesian coordinate system """

    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def __sub__(self, other):
        return CartesianCoordinate(self.x - other.x, self.y - other.y, self.z - other.z)

    def length(self):
        return math.sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2)

    def __str__(self):
        return "<x: " + str(self.x) + ", y: " + str(self.y) + ", z: " + str(self.z) + ">"


class MercatorCoordinate:
    """ Represents a coordinate in Spherical Mercator EPSG:900913 """

    def __init__(self, x=None, y=None):
        self.x = x
        self.y = y

    def to_tile(self, zoom):
        res = _initialresolution / (2 ** zoom)

        def transform(x):
            return int(math.ceil(((x + _originshift) / res) / _tilesize) - 1)

        return TileCoordinate(zoom, transform(self.x),
                              (2 ** zoom - 1) - transform(self.y))

    def to_geographic(self):
        return GeographicCoordinate(
            None if self.x is None else self.x * _180_originshift,
            None if self.y is None
            else (2 * math.atan(math.exp(self.y * _180_originshift * _pi_180)) - _pi_2) \
                 * _180_pi)

    def __str__(self):
        return "<x: " + str(self.x) + ", y: " + str(self.y) + ">"


class MercatorBB:
    """ A bounding box defined by two mercator coordinates """

    def __init__(self, _min=MercatorCoordinate, _max=MercatorCoordinate):
        self.min = _min
        self.max = _max

    def to_tile(self, zoom):
        """ Converts EPSG:900913 to tile coordinates in given zoom level """
        p1 = self.min.to_tile(zoom)
        p2 = self.max.to_tile(zoom)
        return GridBB(zoom, p1.x, p2.y, p2.x, p1.y)

    def __str__(self):
        return "<mercator min: " + str(self.min) + ", max: " + str(self.max) + ">"


class GridCoordinate(metaclass=ABCMeta):
    """
    Represents a position in a worldwide grid.
    """

    def __init__(self, zoom, x, y):
        self.zoom = zoom
        self.x = x
        self.y = y

    def zoom_in(self):
        """
        Yields:
            GridCoordinate: the four tiles of the next zoom level
        """
        return

    def encode_quad_tree(self):
        quad_key = ""
        tx = self.x
        ty = self.y
        for i in range(self.zoom, 0, -1):
            digit = 0
            mask = 1 << (i - 1)
            if (tx & mask) != 0:
                digit += 1
            if (ty & mask) != 0:
                digit += 2
            quad_key += str(digit)
        return quad_key

    def __str__(self):
        return "<zoom: " + str(self.zoom) + ", x: " + str(self.x) + ", y: " + str(self.y) + ">"


class TileCoordinate(GridCoordinate):
    """ Represents a coordinate in a worldwide tile grid. (aka the google-system)"""

    def __init__(self, zoom, x, y):
        super(TileCoordinate, self).__init__(zoom, x, y)

    def to_mercator(self):
        res = _initialresolution / (2 ** self.zoom)
        return MercatorCoordinate(
            None if self.x is None else self.x * _tilesize * res - _originshift,
            None if self.y is None else (2 ** self.zoom - self.y) * _tilesize * res - _originshift)

    def to_geographic(self):
        return self.to_mercator().to_geographic()

    def geographic_bounds(self):
        p1 = self.to_geographic()
        p2 = TileCoordinate(self.zoom, self.x + 1, self.y + 1).to_geographic()
        return GeographicBB(p1.lon, p2.lat, p2.lon, p1.lat)

    def zoom_in(self):
        for x, y in griditer(self.x * 2, self.y * 2, ncol=2):
            yield TileCoordinate(self.zoom + 1, x, y)

    def resolution(self):
        """
        Get the tile resolution at the current position.

        The scale in WG84 depends on
            * the zoom level (obviously)
            * the latitude
            * the tile size

        References:
            * http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Resolution_and_Scale
            * http://gis.stackexchange.com/questions/7430/what-ratio-scales-do-google-maps-zoom-levels-correspond-to

        Returns:
            float: meters per pixel
        """
        geo_coords = self.to_geographic()
        resolution = abs(_initialresolution * math.cos(geo_coords.lat * _pi_180) / (2**self.zoom))
        return resolution


class RegionCoordinate(GridCoordinate):
    """
    Represents a region containing multiple tiles in a worldwide grid of regions.

    A region is a square containing multiple tiles, e.g.::

         -- --
        | 1| 2|
         -- --
        | 3| 4|
         -- --

    A region must contain at least one tile and each row must have
    a power of two of tiles. The size of the region is specified
    with log2(tiles per row per region), e.g.

        * log_tiles_per_row = 0 means 2**0 = 1 tile
        * log_tiles_per_row = 2 means 2**2 = 4 tiles per row, thus 16 tiles per region.

    Args:
        zoom (int): clear.
        x (int): clear.
        y (int): clear.
        log_tiles_per_row (int): size of the region as log2(tiles per row per region).
        needs to be at least 0 (-> 1 tile) and at most 4 (-> 256 tiles)
    """
    def __init__(self, zoom, x, y, log_tiles_per_row=0):
        assert log_tiles_per_row in range(0, 5)
        super().__init__(zoom, x, y)
        self.log_tiles_per_row = log_tiles_per_row
        self.tiles_per_row = 2 ** log_tiles_per_row
        self.root_tile = TileCoordinate(zoom, x * self.tiles_per_row, y * self.tiles_per_row)

    def geographic_bounds(self):
        p1 = self.root_tile.to_geographic()
        p2 = TileCoordinate(self.root_tile.zoom,
                            self.root_tile.x + self.tiles_per_row,
                            self.root_tile.y + self.tiles_per_row).to_geographic()
        return GeographicBB(p1.lon, p2.lat, p2.lon, p1.lat)

    def get_tiles(self):
        """Get all TileCoordinates contained in the region"""
        for x, y in griditer(self.root_tile.x, self.root_tile.y, ncol=self.tiles_per_row):
            yield TileCoordinate(self.root_tile.zoom, x, y)

    def zoom_in(self):
        for x, y in griditer(self.x * 2, self.y * 2, ncol=2):
            yield RegionCoordinate(self.zoom + 1, x, y, log_tiles_per_row=self.log_tiles_per_row)

    def __str__(self):
        return ("<zoom: " + str(self.zoom) + ", x: " + str(self.x) + ", y: " + str(self.y) +
                ", log_tiles_per_row: " + str(self.log_tiles_per_row) + ">")


class GridBB:
    """ A bounding box defined by two grid coordinates """

    def __init__(self, zoom, min_x, min_y, max_x, max_y):
        self.zoom = zoom
        self.min = GridCoordinate(zoom, min_x, min_y)
        self.max = GridCoordinate(zoom, max_x, max_y)

    def intersection(self, other):
        intersects = self.min.x <= other.max.x and self.max.x >= other.min.x and \
                     self.min.y <= other.max.y and self.max.y >= other.min.y
        if intersects:
            return GridBB(self.zoom,
                          max(self.min.x, other.min.x),
                          max(self.min.y, other.min.y),
                          min(self.max.x, other.max.x),
                          min(self.max.y, other.max.y))
        else:
            return []

    def intersections(self, other):
        intersections = []
        i = self.intersection(other)
        if i: intersections.append(i)
        if self.max.x >= 2 ** self.zoom:
            i = GridBB(self.zoom, 0, self.min.y, self.max.x % (2 ** self.zoom), self.max.y).intersection(other)
            if i: intersections.append(i)
        return intersections

    def is_inside(self, tile):
        return self.min.y <= tile.y <= self.max.y and \
               (self.min.x <= tile.x <= self.max.x or
                (self.max.x > 2 ** tile.zoom - 1 and
                 0 <= tile.x <= self.max.x % (2 ** tile.zoom)))

    def __str__(self):
        return "<tile min: " + str(self.min) + ", max: " + str(self.max) + ">"



""" pyKML Utility Module

The pykml.utility module provides utility functions that operate on KML 
documents
"""
import re

def count_elements(doc):
    "Counts the number of times each element is used in a document"
    summary = {}
    for el in doc.iter():
        try:
            namespace, element_name = re.search('^{(.+)}(.+)$', el.tag).groups()
        except:
            namespace = None
            element_name = el.tag
        if not summary.has_key(namespace):
            summary[namespace] = {}
        if not summary[namespace].has_key(element_name):
            summary[namespace][element_name] = 1
        else:
            summary[namespace][element_name] += 1
    return summary



"""
Module for printing maps / exporting them as pdf
"""

from geos.geometry import *
import urllib.request
from urllib.error import URLError
from PIL import Image, ImageDraw
from geos import app
from tempfile import NamedTemporaryFile
from multiprocessing import Pool
import pkg_resources

TILE_SIZE = 256  # px
# some tile servers block requests noch coming from a Browser
USER_AGENT = 'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, ' \
             'like Gecko) Chrome/35.0.1916.153 Safari/537.36 SE 2.X MetaSr 1.0'
N_DOWNLOAD_WORKERS = 16

# set default user agent
opener = urllib.request.build_opener()
opener.addheaders = [('User-agent', USER_AGENT)]
urllib.request.install_opener(opener)


class MapPrintError(Exception):
    pass


def dpi_to_dpmm(dpi):
    """
    Convert dpi (dots per inch) to dpmm (dots per millimeter)

    Args:
        dpi (int):

    Returns:
        float: dots per millimeter
    """
    return dpi / 25.4


def print_map(map_source, x, y, zoom=14, width=297, height=210, dpi=300, format="pdf"):
    """
    Download map tiles and stitch them together in a single image, ready for printing.

    Args:
        map_source (MapSource): Map to download
        x (float): map center x-coordinate in Mercator projection (EPSG:4326)
        y (float): map center y-coordinate in Mercator projection (EPSG:4326)
        zoom (int): tile zoom level to use for printing
        width (float): page width in mm
        height (float): page height in mm
        dpi (int): resolution in dots per inch
        format (str): output format. Anything supported by ``Pillow.Image.save``. E.g. "pdf", "jpeg", "png".

    Returns:
        str: path of temporary output file.

    """
    bbox = get_print_bbox(x, y, zoom, width, height, dpi)
    tiles = [
        get_tiles(tile_layer, bbox) for tile_layer in map_source.layers if
        tile_layer.min_zoom <= zoom <= tile_layer.max_zoom
        ]
    img = stitch_map(tiles, width, height, bbox, dpi)
    outfile = NamedTemporaryFile(delete=False)
    img.save(outfile, format, quality=100, dpi=(dpi, dpi))
    outfile.close()
    return outfile.name


def get_print_bbox(x, y, zoom, width, height, dpi):
    """
    Calculate the tile bounding box based on position, map size and resolution.

    The function returns the next larger tile-box, that covers the specified
    page size in mm.

    Args:
        x (float): map center x-coordinate in Mercator projection (EPSG:4326)
        y (float): map center y-coordinate in Mercator projection (EPSG:4326)
        zoom (int): tile zoom level to use for printing
        width (float): page width in mm
        height (float): page height in mm
        dpi (int): resolution in dots per inch

    Returns:
        GridBB: Bounding box of the map in TileCoordinates.

    >>> str(get_print_bbox(4164462.1505763642, 985738.7965919945, 14, 297, 150, 120))
    '<tile min: <zoom: 14, x: 9891, y: 7786>, max: <zoom: 14, x: 9897, y: 7790>>'
    """
    tiles_h = width * dpi_to_dpmm(dpi) / TILE_SIZE
    tiles_v = height * dpi_to_dpmm(dpi) / TILE_SIZE
    mercator_coords = MercatorCoordinate(x, y)
    tile_coords = mercator_coords.to_tile(zoom)
    tile_bb = GridBB(zoom,
                     min_x=tile_coords.x - math.ceil(tiles_h / 2),
                     max_x=tile_coords.x + math.ceil(tiles_h / 2),
                     min_y=tile_coords.y - math.ceil(tiles_v / 2),
                     max_y=tile_coords.y + math.ceil(tiles_v / 2))
    return tile_bb


def download_tile(map_layer, zoom, x, y):
    """
    Download a given tile from the tile server.

    Args:
        map_layer (MapLayer): MapLayer object which provides the tile-url.
        zoom (int): zoom level
        x (int): Tile-x-coordinate
        y (int): Tile-y-coordinate

    Returns:
        file: temporary file containing the downloaded image.

    """
    try:
        tile_url = map_layer.get_tile_url(zoom, x, y)
        tmp_file, headers = urllib.request.urlretrieve(tile_url)
        return (x, y), tmp_file
    except URLError as e:
        app.logger.info("Error downloading tile x={}, y={}, z={} for layer {}: {}".format(
            x, y, zoom, map_layer, e.reason))
        return (x, y), pkg_resources.resource_filename("geos", "static/empty_tile.png")


def _download_tile_wrapper(args):
    """wrapper function for using the multiprocessing module with multiple arguments"""
    return download_tile(*args)


def get_tiles(map_layer, bbox, n_workers=N_DOWNLOAD_WORKERS):
    """
    Download tiles.

    Args:
        map_source (MapSource):
        bbox (TileBB): Bounding box delimiting the map
        n_workers (int): number of threads to used for downloading.

    Returns:
        dict of file: Dictionary mapping coordinates to temporary files.

        Example::

            {
                (x, y) : <FileHandle>
            }

    """
    p = Pool(n_workers)
    tiles = {}
    for (x, y), tmp_file in p.imap_unordered(_download_tile_wrapper, zip(itertools.repeat(map_layer),
                                                                         itertools.repeat(bbox.zoom),
                                                                         *zip(*bboxiter(bbox)))):
        app.logger.info("Downloaded tile x={}, y={}, z={}".format(x, y, bbox.zoom))
        tiles[(x, y)] = tmp_file

    return tiles


def stitch_map(tiles, width, height, bbox, dpi):
    """
    Merge tiles together into one image.

    Args:
        tiles (list of dict of file): tiles for each layer
        width (float): page width in mm
        height (height): page height in mm
        dpi (dpi): resolution in dots per inch

    Returns:
        PIL.Image: merged map.

    """
    size = (int(width * dpi_to_dpmm(dpi)), int(height * dpi_to_dpmm(dpi)))
    background = Image.new('RGBA', size, (255, 255, 255))
    for layer in tiles:
        layer_img = Image.new("RGBA", size)
        for (x, y), tile_path in layer.items():
            tile = Image.open(tile_path)
            layer_img.paste(tile, ((x - bbox.min.x) * TILE_SIZE, (y - bbox.min.y) * TILE_SIZE))
        background = Image.alpha_composite(background, layer_img)
    add_scales_bar(background, bbox)
    return background.convert("RGB")


def add_scales_bar(img, bbox):
    """
    Add a scales bar to the  map.

    Calculates the resolution at the current latitude and
    inserts the corresponding scales bar on the map.

    Args:
        img (Image): Image object to which the scales bar will be added.
        bbox (TileBB): boundaries of the map

    """
    tc = TileCoordinate(bbox.min.zoom, bbox.min.x, bbox.min.y)
    meters_per_pixel = tc.resolution()
    one_km_bar = int(1000 * (1 / meters_per_pixel))
    col_black = (0, 0, 0)

    line_start = (100, img.size[1] - 100)  # px
    line_end = (line_start[0] + one_km_bar, line_start[1])
    whiskers_left = [line_start[0], line_start[1] - 15, line_start[0], line_start[1] + 15]
    whiskers_right = [line_end[0], line_end[1] - 15, line_end[0], line_end[1] + 15]

    draw = ImageDraw.Draw(img)
    draw.line([line_start, line_end], fill=col_black, width=5)
    draw.line(whiskers_left, fill=col_black, width=2)
    draw.line(whiskers_right, fill=col_black, width=2)
    draw.text((line_start[0] + 10, line_start[1] + 10), fill=col_black, text="1 km")
    del draw



'''pyKML Factory Module

The pykml.factory module provides objects and functions that can be used to 
create KML documents element-by-element. 
The factory module leverages `lxml's ElementMaker factory`_ objects to create
KML objects with the appropriate namespace prefixes.

.. _lxml: http://lxml.de
.. _lxml's ElementMaker factory: http://lxml.de/objectify.html#tree-generation-with-the-e-factory
'''

from lxml import etree, objectify

nsmap={
    None: "http://www.opengis.net/kml/2.2",
    'atom': "http://www.w3.org/2005/Atom",
    'gx': "http://www.google.com/kml/ext/2.2",
}

# create a factory object for creating objects in the KML namespace
KML_ElementMaker = objectify.ElementMaker(
    annotate=False,
    namespace=nsmap[None],
    nsmap=nsmap,
)
# create a factory object for creating objects in the ATOM namespace
ATOM_ElementMaker = objectify.ElementMaker(
    annotate=False,
    namespace=nsmap['atom'],
    nsmap={'atom': nsmap['atom']},
)
# Create a factory object for the KML Google Extension namespace
GX_ElementMaker = objectify.ElementMaker(
    annotate=False,
    namespace=nsmap['gx'],
    nsmap={'gx': nsmap['gx']},
)

def get_factory_object_name(namespace):
    "Returns the correct factory object for a given namespace"
    
    factory_map = {
        'http://www.opengis.net/kml/2.2': 'KML',
        'http://www.w3.org/2005/Atom': 'ATOM',
        'http://www.google.com/kml/ext/2.2': 'GX'
    }
    if namespace:
        if factory_map.has_key(namespace):
            factory_object_name = factory_map[namespace]
        else:
            factory_object_name = None
    else:
        # use the KML factory object as the default, if no namespace is given
        factory_object_name = 'KML'
    return factory_object_name

def write_python_script_for_kml_document(doc):
    "Generates a python script that will construct a given KML document"
    import StringIO
    from pykml.helpers import separate_namespace
    
    output = StringIO.StringIO()
    indent_size = 2
    
    # add the etree package so that comments can be handled
    output.write('from lxml import etree\n')
    
    # add the namespace declaration section
    output.write('from pykml.factory import KML_ElementMaker as KML\n')
    output.write('from pykml.factory import ATOM_ElementMaker as ATOM\n')
    output.write('from pykml.factory import GX_ElementMaker as GX\n')
    output.write('\n')
    
    level = 0
    xml = StringIO.StringIO(etree.tostring(doc))
    context = etree.iterparse(xml, events=("start", "end", "comment"))
    output.write('doc = ')
    last_action = None
    main_element_processed_flag = False
    previous_list = []  # list of comments before the root element
    posterior_list = []  # list of comments after the root element
    for action, elem in context:
        # TODO: remove the following redundant conditional
        if action in ('start','end','comment'):
            namespace, element_name = separate_namespace(elem.tag)
            if action in ('comment'):
                indent = ' ' * level * indent_size
                if elem.text:
                    text_list = elem.text.split('\n')
                    if len(text_list) == 1:
                        text = repr(elem.text)
                    else: # len(text_list) > 1
                        # format and join all non-empty lines
                        text = '\n' + ''.join(
                                ['{indent}{content}\n'.format(
                                        indent=' '*(len(t)-len(t.lstrip())),
                                        content=repr(t.lstrip())
                                    ) for t in text_list if len(t.strip())>0
                                ]
                            ) + indent
                else:
                    text = ''
                if level == 0:
                    # store the comment so that it can be appended later
                    if main_element_processed_flag:
                        posterior_list.append("{indent}etree.Comment({comment})".format(
                            indent = indent,
                            comment = text,
                        ))
                    else:
                        previous_list.append("{indent}etree.Comment({comment})".format(
                            indent = indent,
                            comment = text,
                        ))
                else:
                    output.write("{indent}etree.Comment({comment}),\n".format(
                        indent = indent,
                        comment = text,
                    ))
            elif action in ('start'):
                main_element_processed_flag = True
                if last_action == None:
                    indent = ''
                else:
                    indent = ' ' * level * indent_size
                level += 1
                if elem.text:
                    # METHOD 1
#                    # treat all text string the same. this works but gets
#                    # messy for multi-line test strings
#                    text = repr(elem.text)
                    # METHOD 2 - format multiline strings differently
                    text_list = elem.text.split('\n')
                    if len(text_list) == 1:
                        text = repr(elem.text)
                    else: # len(text_list) > 1
                        # format and join all non-empty lines
                        text = '\n' + ''.join(
                                ['{indent}{content}\n'.format(
                                        indent=' '*(len(t)-len(t.lstrip())),
                                        content=repr(t.lstrip())
                                    ) for t in text_list if len(t.strip())>0
                                ]
                            ) + indent
                else:
                    text = ''
                output.write('{indent}{factory}.{tag}({text}\n'.format(
                    indent = indent,
                    factory = get_factory_object_name(namespace),
                    tag = element_name,
                    text = text,
                ))
            elif action in ('end'):
                level -= 1
                if last_action == 'start':
                    output.pos -= 1
                    indent = ''
                else:
                    indent = ' ' * level * indent_size
                for att,val in elem.items():
                    output.write('{0}  {1}="{2}",\n'.format(indent,att,val))
                output.write('{0}),\n'.format(indent))
        last_action = action
    
    # remove the last comma
    output.pos -= 2
    output.truncate()
    output.write('\n')
    
    for entry in previous_list:
        output.write('doc.addprevious({entry})\n'.format(
            entry=entry
        ))
    for entry in posterior_list:
        output.write('doc.addnext({entry})\n'.format(
            entry=entry
        ))
    
    # add python code to print out the KML document
    output.write('print etree.tostring(etree.ElementTree(doc),pretty_print=True)\n')
    
    return output.getvalue()


#!/usr/bin/env python3

"""
Start the GEOS server from command line.
"""

from geos import app
import argparse
import geos.mapsource
import pkg_resources
from geos.kml import URLFormatter

MAPSOURCES = pkg_resources.resource_filename("geos", "mapsources")

app.config.from_object('geos.default_settings')


def run_app(default_host=app.config['HOST'], default_port=app.config['PORT']):
    argp = argparse.ArgumentParser("geos")
    argp.add_argument("-m", "--mapsource", required=False,
                      default=None,
                      help="path to the directory containing the mapsource files. [default: integrated mapsources]")
    argp.add_argument("-H", "--host", required=False,
                      help="Hostname of the Flask app [default {}]".format(default_host),
                      default=default_host)
    argp.add_argument("-P", "--port", required=False,
                      help="Port for the Flask app [default {}]".format(default_port),
                      default=default_port)
    argp.add_argument("--display-host", required=False,
                      help="Hostname used for self-referencing links [defaults to Flask hostname]",
                      default=None)
    argp.add_argument("--display-port", required=False,
                      help="Port used for self-referencing links [defaults to Flask port]",
                      default=None)
    argp.add_argument("--display-scheme", required=False,
                      help="URI-scheme used for self-referencing links [default {}]".format(app.config["PREFERRED_URL_SCHEME"]),
                      default=None)

    args = argp.parse_args()

    app.config['url_formatter'] = URLFormatter(
        host=args.display_host if args.display_host else args.host,
        port=args.display_port if args.display_port else args.port,
        url_scheme=args.display_scheme if args.display_scheme else app.config["PREFERRED_URL_SCHEME"]
    )
    app.config['mapsources'] = geos.mapsource.load_maps(MAPSOURCES)
    if args.mapsource is not None:
        app.config["mapsources"].update(geos.mapsource.load_maps(args.mapsource))

    app.run(
        host=args.host,
        port=int(args.port)
    )

if __name__ == "__main__":
    run_app()



import unittest
from os import path
import urllib2
from StringIO import StringIO
from lxml import etree
from pykml.parser import Schema
from pykml.parser import fromstring
from pykml.parser import parse


class ValidatorTestCase(unittest.TestCase):
    
    def test_initialize_schema(self):
        """Tests the creation Schema instance"""
        schema = Schema("ogckml22.xsd")
        self.assertTrue(isinstance(schema.schema, etree.XMLSchema))


class ParseKmlOgcTestCase(unittest.TestCase):
    "A collection of tests related to parsing KML OGC documents"
    
    def test_fromstring_kml_document(self):
        "Tests the parsing of an valid KML string"
        test_kml = '<kml xmlns="http://www.opengis.net/kml/2.2"/>'
        tree = fromstring(test_kml, schema=Schema("ogckml22.xsd"))
        self.assertEquals(etree.tostring(tree), test_kml)
        tree = fromstring(test_kml)
        self.assertEquals(etree.tostring(tree), test_kml)
    
    def test_fromstring_invalid_kml_document(self):
        "Tests the parsing of an invalid KML string"
        test_kml = '<bad_element />'
        try:
            tree = fromstring(test_kml, schema=Schema("ogckml22.xsd"))
            self.assertTrue(False)
        except etree.XMLSyntaxError:
            self.assertTrue(True)
        except:
            self.assertTrue(False)
    
    def test_parse_kml_document(self):
        "Tests the parsing of an valid KML file object"
        test_kml = '<kml xmlns="http://www.opengis.net/kml/2.2"/>'
        fileobject = StringIO(test_kml)
        schema = Schema("ogckml22.xsd")
        tree = parse(fileobject, schema=schema)
        self.assertEquals(etree.tostring(tree), test_kml)
        tree = parse(fileobject, schema=schema)
        self.assertEquals(etree.tostring(tree), test_kml)
    
    def test_parse_invalid_kml_document(self):
        "Tests the parsing of an invalid KML document"
        fileobject = StringIO('<bad_element />')
        try:
            tree = parse(fileobject, schema=Schema("ogckml22.xsd"))
            self.assertTrue(False)
        except etree.XMLSyntaxError:
            self.assertTrue(True)
        except:
            self.assertTrue(False)
    
    def test_parse_kml_url(self):
        "Tests the parsing of a KML URL"
        url = 'http://code.google.com/apis/kml/documentation/KML_Samples.kml'
        #url = 'http://kml-samples.googlecode.com/svn/trunk/kml/Document/doc-with-id.kml'
        #url = 'http://code.google.com/apis/kml/documentation/kmlfiles/altitudemode_reference.kml'
        #url = 'http://code.google.com/apis/kml/documentation/kmlfiles/animatedupdate_example.kml'
        fileobject = urllib2.urlopen(url)
        tree = parse(fileobject, schema=Schema("ogckml22.xsd"))
        self.assertEquals(
            etree.tostring(tree)[:78],
            '<kml xmlns="http://www.opengis.net/kml/2.2">'
              '<Document>'
                '<name>KML Samples</name>'
        )
    
    def test_parse_invalid_ogc_kml_document(self):
        """Tests the parsing of an invalid KML document.  Note that this KML
        document uses elements that are not in the OGC KML spec.
        """
        url = 'http://code.google.com/apis/kml/documentation/kmlfiles/altitudemode_reference.kml'
        fileobject = urllib2.urlopen(url)
        try:
            tree = parse(fileobject, schema=Schema("ogckml22.xsd"))
            self.assertTrue(False)
        except etree.XMLSyntaxError:
            self.assertTrue(True)
        except:
            self.assertTrue(False)

class ParseKmlGxTestCase(unittest.TestCase):
    "A collection of tests related to parsing KML Google Extension documents"
    
    def test_parse_kml_url(self):
        "Tests the parsing of a KML URL"
        url = 'http://code.google.com/apis/kml/documentation/kmlfiles/altitudemode_reference.kml'
        fileobject = urllib2.urlopen(url)
        tree = parse(fileobject, schema=Schema('kml22gx.xsd'))
        self.assertEquals(
            etree.tostring(tree)[:185],
            '<kml xmlns="http://www.opengis.net/kml/2.2" '
                 'xmlns:gx="http://www.google.com/kml/ext/2.2">'
                '<!-- required when using gx-prefixed elements -->'
                '<Placemark>'
                  '<name>gx:altitudeMode Example</name>'
        )
    
    def test_parse_kml_file(self):
        "Tests the parsing of a local KML file, with validation"
        test_datafile = path.join(
            path.dirname(__file__),
            'testfiles',
            'google_kml_developers_guide/complete_tour_example.kml'
        )
        # parse with validation
        with open(test_datafile) as f:
            doc = parse(f, schema=Schema('kml22gx.xsd'))
        # parse without validation
        with open(test_datafile) as f:
            doc = parse(f)
        self.assertTrue(True)
    
    def test_parse_kml_file_with_cdata(self):
        "Tests the parsing of a local KML file, with a CDATA description string"
        test_datafile = path.join(
            path.dirname(__file__),
            'testfiles',
            'google_kml_tutorial/using_the_cdata_element.kml'
        )
        # parse with validation
        with open(test_datafile) as f:
            doc = parse(f, schema=Schema('kml22gx.xsd'))
        self.assertEquals(
            etree.tostring(doc),
            '<kml xmlns="http://www.opengis.net/kml/2.2">'
              '<Document>'
                '<Placemark>'
                  '<name>CDATA example</name>'
                  '<description>'
                    '<![CDATA[\n'
                    '          <h1>CDATA Tags are useful!</h1>\n'
                    '          <p><font color="red">Text is <i>more readable</i> and \n'
                    '          <b>easier to write</b> when you can avoid using entity \n'
                    '          references.</font></p>\n'
                    '        ]]>'
                  '</description>'
                '<Point>'
                  '<coordinates>102.595626,14.996729</coordinates>'
                '</Point>'
              '</Placemark>'
            '</Document>'
          '</kml>'
        )
    
    def test_parse_kml_url_2(self):
        "Tests the parsing of a KML URL"
        url = 'http://code.google.com/apis/kml/documentation/kmlfiles/animatedupdate_example.kml'
        fileobject = urllib2.urlopen(url)
        tree = parse(fileobject, schema=Schema('kml22gx.xsd'))
        self.assertEquals(
            etree.tostring(tree)[:137],
            '<kml xmlns="http://www.opengis.net/kml/2.2" '
                 'xmlns:gx="http://www.google.com/kml/ext/2.2">'
                '<Document>'
                  '<name>gx:AnimatedUpdate example</name>'
        )

if __name__ == '__main__':
    unittest.main()


import xml.etree.ElementTree
import itertools
import os
from pprint import pprint
from geos.geometry import GeographicBB
import re
import random

F_SEP = "/"  # folder separator in mapsources (not necessarily == os.sep)


def load_maps(maps_dir):
    """
    Load all xml map sources from a given directory.

    Args:
        maps_dir: path to directory to search for maps

    Returns:
        dict of MapSource:

    """
    maps_dir = os.path.abspath(maps_dir)
    maps = {}
    for root, dirnames, filenames in os.walk(maps_dir):
        for filename in filenames:
            if filename.endswith(".xml"):
                xml_file = os.path.join(root, filename)
                map = MapSource.from_xml(xml_file, maps_dir)
                if map.id in maps:
                    raise MapSourceException("duplicate map id: {} in file {}".format(map.id, xml_file))
                else:
                    maps[map.id] = map
    return maps


def walk_mapsources(mapsources, root=""):
    """
    recursively walk through foldernames of mapsources.

    Like os.walk, only for a list of mapsources.

    Args:
        mapsources (list of MapSource):

    Yields:
        (root, foldernames, maps)

    >>> mapsources = load_maps("test/mapsources")
    >>> pprint([x for x in walk_mapsources(mapsources.values())])
    [('',
      ['asia', 'europe'],
      [<MapSource: osm1 (root 1), n_layers: 1, min_zoom:5, max_zoom:18>,
       <MapSource: osm10 (root 2), n_layers: 1, min_zoom:5, max_zoom:18>]),
     ('/asia',
      [],
      [<MapSource: osm6 (asia), n_layers: 1, min_zoom:5, max_zoom:18>]),
     ('/europe',
      ['france', 'germany', 'switzerland'],
      [<MapSource: osm4 (eruope 1), n_layers: 1, min_zoom:1, max_zoom:18>]),
     ('/europe/france',
      [],
      [<MapSource: osm2 (europe/france 1), n_layers: 1, min_zoom:5, max_zoom:18>,
       <MapSource: osm3 (europe/france 2), n_layers: 1, min_zoom:1, max_zoom:18>,
       <MapSource: osm5 (europe/france 3), n_layers: 1, min_zoom:5, max_zoom:18>]),
     ('/europe/germany',
      [],
      [<MapSource: osm7 (europe/germany 1), n_layers: 1, min_zoom:5, max_zoom:18>,
       <MapSource: osm8 (europe/germany 2), n_layers: 1, min_zoom:5, max_zoom:18>]),
     ('/europe/switzerland',
      [],
      [<MapSource: osm9 (europe/switzerland), n_layers: 1, min_zoom:5, max_zoom:18>])]

    """
    def get_first_folder(path):
        """
        Get the first folder in a path
        > get_first_folder("europe/switzerland/bs")
        europe
        """
        path = path[len(root):]
        path = path.lstrip(F_SEP)
        return path.split(F_SEP)[0]

    path_tuples = sorted(((get_first_folder(m.folder), m) for m in mapsources), key=lambda x: x[0])
    groups = {k: [x for x in g] for k, g in itertools.groupby(path_tuples, lambda x: x[0])}
    folders = sorted([x for x in groups.keys() if x != ""])
    mapsources = sorted([t[1] for t in groups.get("", [])], key=lambda x: x.id)
    yield (root, folders, mapsources)
    for fd in folders:
        yield from walk_mapsources([t[1] for t in groups[fd]], F_SEP.join([root, fd]))


class MapSourceException(Exception):
    pass


class MapLayer(object):
    """
    The layer object contained in a MapSource.

    A MapSource can contain multiple layers.
    A layer contains all information on how to access the tiles.

    Args:
        tile_url: URL to the tiles which {$z}, {$x} and {$y} as placeholders.
        min_zoom (int): minimal zoom level at which the layer is active
        max_zoom (int): maximal zoom level at which the layer is active
    """
    def __init__(self, tile_url=None, min_zoom=1, max_zoom=17):
        self.tile_url = tile_url
        self.min_zoom = min_zoom
        self.max_zoom = max_zoom
        self.server_parts = []

    def get_tile_url(self, zoom, x, y):
        """
        Fill the placeholders of the tile url with zoom, x and y.

        >>> ms = MapSource.from_xml("mapsources/osm.xml")
        >>> ms.layers[0].get_tile_url(42, 43, 44)
        'http://tile.openstreetmap.org/42/43/44.png'
        """
        if(len(self.server_parts) == 0):
            return self.tile_url.format(**{"$z": zoom, "$x": x, "$y": y})
        else:
            return self.tile_url.format(**{"$z": zoom, "$x": x, "$y": y, "$serverpart": random.choice(self.server_parts)})

    @property
    def get_tile_urls(self):
        if(len(self.server_parts) == 0):
            return [self.tile_url.replace("$", "")]
        else:
            cur_tile_url = []
            for s in self.server_parts:
                cur_tile_url.append(self.tile_url.replace("{$serverpart}", s).replace("$", ""))
        return cur_tile_url
        
    def __repr__(self):
        return "<MapLayer: url:{}, min_zoom:{}, max_zoom:{}>".format(
            self.tile_url, self.min_zoom, self.max_zoom)


class MapSource(object):
    """
    A MapSource contains Meta-Information of the map.
    Additionally it can hold one or more MapLayers which contain the information
    on how to access the tiles.
    """

    def __init__(self, id, name, folder="", bbox=None):
        """
        Args:
            id (str): unique identifier (e.g. filename)
            name (str): display name
            folder (str): folder to organize the map in (e.g. /europe/germany)
            bbox (GeographicBB): bounding box, only load tiles within

        >>> ms = MapSource.from_xml("mapsources/osm.xml")
        >>> ms.name
        'OSM Mapnik (default)'
        >>> ms.min_zoom
        0
        >>> ms.max_zoom
        18
        >>> ms.layers
        [<MapLayer: url:http://tile.openstreetmap.org/{$z}/{$x}/{$y}.png, min_zoom:0, max_zoom:18>]
        """
        self.id = id
        self.name = name
        self.layers = []
        self.folder = folder
        self.bbox = bbox

    @property
    def min_zoom(self):
        """
        Get the minimal zoom level of all layers.

        Returns:
            int: the minimum of all zoom levels of all layers

        Raises:
            ValueError: if no layers exist

        """
        zoom_levels = [map_layer.min_zoom for map_layer in self.layers]
        return min(zoom_levels)

    @property
    def max_zoom(self):
        """
        Get the maximal zoom level of all layers.

        Returns:
            int: the maximum of all zoom levels of all layers

        Raises:
            ValueError: if no layers exist

        """
        zoom_levels = [map_layer.max_zoom for map_layer in self.layers]
        return max(zoom_levels)

    @staticmethod
    def parse_xml_boundary(xml_region):
        """
        Get the geographic bounds from an XML element

        Args:
            xml_region (Element): The <region> tag as XML Element

        Returns:
            GeographicBB:
        """
        try:
            bounds = {}
            for boundary in xml_region:
                bounds[boundary.tag] = float(boundary.text)
            bbox = GeographicBB(min_lon=bounds["west"], max_lon=bounds["east"],
                                min_lat=bounds["south"], max_lat=bounds["north"])
            return bbox
        except (KeyError, ValueError):
            raise MapSourceException("region boundaries are invalid. ")

    @staticmethod
    def parse_xml_layers(xml_layers):
        """
        Get the MapLayers from an XML element

        Args:
            xml_layers (Element): The <layers> tag as XML Element

        Returns:
            list of MapLayer:

        """
        layers = []
        for custom_map_source in xml_layers:
            layers.append(MapSource.parse_xml_layer(custom_map_source))
        return layers

    @staticmethod
    def parse_xml_layer(xml_custom_map_source):
        """
        Get one MapLayer from an XML element

        Args:
            xml_custom_map_source (Element): The <customMapSource> element tag wrapped
               in a <layers> tag as XML Element

        Returns:
            MapLayer:

        """
        map_layer = MapLayer()
        try:
            for elem in xml_custom_map_source.getchildren():
                if elem.tag == 'url':
                    map_layer.tile_url = elem.text
                elif elem.tag == 'minZoom':
                    map_layer.min_zoom = int(elem.text)
                elif elem.tag == 'maxZoom':
                    map_layer.max_zoom = int(elem.text)
                elif elem.tag == 'serverParts':
                    map_layer.server_parts += str(elem.text).split()
        except ValueError:
            raise MapSourceException("minZoom/maxZoom must be an integer. ")

        if map_layer.tile_url is None:
            raise MapSourceException("Layer requires a tile_url parameter. ")

        return map_layer

    @staticmethod
    def from_xml(xml_path, mapsource_prefix=""):
        """
        Create a MapSource object from a MOBAC
        mapsource xml.

        Args:
            xml_path: path to the MOBAC mapsource xml file.
            mapsource_prefix: root path of the mapsource folder.
              Used to determine relative path within the maps
              directory.

        Note:
            The Meta-Information is read from the xml
            <id>, <folder>, <name> tags. If <id> is not available it defaults
            to the xml file basename. If <folder> is not available if defaults to
            the folder of the xml file with the `mapsource_prefix` removed.

            The function first tries <url>, <minZoom>, <maxZoom> from <customMapSource>
            tags within the <layers> tag. If the <layers> tag is not available,
            the function tries to find <url>, <minZoom> and <maxZoom> on the top level.
            If none of thie information is found, a MapSourceException is raised. 

        Returns:
            MapSource:

        Raises:
            MapSourceException: when the xml file could not be parsed properly.

        """
        xmldoc = xml.etree.ElementTree.parse(xml_path).getroot()

        map_id = os.path.splitext(os.path.basename(xml_path))[0]
        map_name = map_id
        map_folder = re.sub("^" + re.escape(mapsource_prefix), "", os.path.dirname(xml_path))
        bbox = None
        layers = None

        for elem in xmldoc:
            if elem.tag == 'id':
                map_id = elem.text
            elif elem.tag == 'name':
                map_name = elem.text
            elif elem.tag == 'folder':
                map_folder = elem.text
            elif elem.tag == 'region':
                bbox = MapSource.parse_xml_boundary(elem)
            elif elem.tag == 'layers':
                layers = MapSource.parse_xml_layers(elem)

        if map_folder is None:
            map_folder = "" # fallback if bad specification in xml
        if layers is None:  # layers tag not found, expect url etc. in main xmldoc
            layers = [MapSource.parse_xml_layer(xmldoc)]

        ms = MapSource(map_id, map_name, map_folder, bbox=bbox)
        ms.layers = layers
        return ms

    def __repr__(self):
        return "<MapSource: {} ({}), n_layers: {}, min_zoom:{}, max_zoom:{}>".format(
            self.id, self.name, len(self.layers), self.min_zoom, self.max_zoom)





import unittest
from pykml.factory import KML_ElementMaker as K
from pykml.factory import ATOM_ElementMaker as ATOM
from pykml.factory import GX_ElementMaker as GX
from pykml.parser import Schema
from pykml.parser import fromstring

class KmlHelpersTestCase(unittest.TestCase):
    
    def test_separate_namespace(self):
        """Tests the function that separates namespaces from qualified names"""
        from pykml.helpers import separate_namespace
        
        namespace, element_name = separate_namespace('{garden}eggplant')
        self.assertEqual(namespace, 'garden')
        self.assertEqual(element_name, 'eggplant')
        
        namespace, element_name = separate_namespace('eggplant')
        self.assertEqual(namespace, None)
        self.assertEqual(element_name, 'eggplant')
    
    def test_set_max_decimal_places(self):
        """Tests setting the number of decimal places in a document"""
        
        from pykml.helpers import set_max_decimal_places
        
        test_kml = (
            '<?xml version="1.0" encoding="UTF-8"?>'
            '<kml xmlns="http://www.opengis.net/kml/2.2" '
                 'xmlns:gx="http://www.google.com/kml/ext/2.2" '
                 'xmlns:kml="http://www.opengis.net/kml/2.2" '
                 'xmlns:atom="http://www.w3.org/2005/Atom">'
                '<Document>'
                    '<Placemark>'
                        '<name>Spearhead</name>'
                        '<LookAt>'
                            '<longitude>-105.6381333137406</longitude>'
                            '<latitude>40.25542364754504</latitude>'
                            '<altitude>0.123456789</altitude>'
                            '<heading>-75.2679217880259</heading>'
                            '<tilt>23.33768008163944</tilt>'
                            '<range>234.1234567890</range>'
                            '<altitudeMode>relativeToGround</altitudeMode>'
                        '</LookAt>'
                        '<Point>'
                            '<altitudeMode>absolute</altitudeMode>'
                            '<coordinates>-105.6381333137406,40.25542364754504,3826.1234567890</coordinates>'
                        '</Point>'
                    '</Placemark>'
                    '<Placemark>'
                        '<name>North Ridge</name>'
                        '<LineString>'
                            '<tessellate>1</tessellate>'
                            '<coordinates>'
                                '-105.6400899274733,40.25778038346723,0 -105.6397083557171,40.25680995109639,0 -105.6389193178716,40.25606911151127,0 -105.6385344464865,40.25559941617504,0 '
                            '</coordinates>'
                        '</LineString>'
                    '</Placemark>'
                    '<Placemark>'
                        '<name>Black Lake</name>'
                        '<Polygon>'
                            '<tessellate>1</tessellate>'
                            '<outerBoundaryIs>'
                                '<LinearRing>'
                                    '<coordinates>'
                                        '-105.6411566922825,40.26642751482452,0 -105.6424707115566,40.26600581412304,0 -105.6428010853679,40.26549238687588,0 -105.6423465680717,40.26464873270425,0 -105.6414735029203,40.2645201246226,0 -105.6404885591011,40.2647647504581,0 -105.6402035173736,40.26539046743765,0 -105.6405618656529,40.26617884849238,0 -105.6411566922825,40.26642751482452,0 '
                                    '</coordinates>'
                                '</LinearRing>'
                            '</outerBoundaryIs>'
                        '</Polygon>'
                    '</Placemark>'
                '</Document>'
            '</kml>'
        )
        doc = fromstring(test_kml, schema=Schema("ogckml22.xsd"))
        set_max_decimal_places(
            doc, 
            max_decimals={
                'longitude': 6,
                'latitude': 5,
                'altitude': 2,
                'heading': 1,
                'tilt': 0,
                #'range': 0,  # range values will not be changed
            }
        )
        
        longitude_list = doc.findall(".//{http://www.opengis.net/kml/2.2}longitude")
        self.assertAlmostEquals(longitude_list[0], -105.638133)
        
        latitude_list = doc.findall(".//{http://www.opengis.net/kml/2.2}latitude")
        self.assertAlmostEquals(latitude_list[0], 40.25542)
        
        altitude_list = doc.findall(".//{http://www.opengis.net/kml/2.2}altitude")
        self.assertAlmostEquals(altitude_list[0], 0.12)
        
        heading_list = doc.findall(".//{http://www.opengis.net/kml/2.2}heading")
        self.assertAlmostEquals(heading_list[0], -75.3)
        
        tilt_list = doc.findall(".//{http://www.opengis.net/kml/2.2}tilt")
        self.assertAlmostEquals(tilt_list[0], 23.0)
        
        # Note that the range value was not changed
        range_list = doc.findall(".//{http://www.opengis.net/kml/2.2}range")
        self.assertAlmostEquals(range_list[0], 234.1234567890)
        
        coords_list = doc.findall(".//{http://www.opengis.net/kml/2.2}coordinates")
        self.assertEquals(
            coords_list[0],
            "-105.638133,40.25542,3826.12"
        )
        self.assertEquals(
            coords_list[1], 
            '-105.64009,40.25778,0.0 '
            '-105.639708,40.25681,0.0 '
            '-105.638919,40.25607,0.0 '
            '-105.638534,40.2556,0.0'
        )
        #import ipdb; ipdb.set_trace()
        self.assertEquals(
            coords_list[2], 
            '-105.641157,40.26643,0.0 '
            '-105.642471,40.26601,0.0 '
            '-105.642801,40.26549,0.0 '
            '-105.642347,40.26465,0.0 '
            '-105.641474,40.26452,0.0 '
            '-105.640489,40.26476,0.0 '
            '-105.640204,40.26539,0.0 '
            '-105.640562,40.26618,0.0 '
            '-105.641157,40.26643,0.0'
        )


    def test_set_max_decimal_places_track(self):
        """Tests setting the number of decimal places for track data"""
        
        from pykml.helpers import set_max_decimal_places
        
        test_kml = (
            '<?xml version="1.0" encoding="UTF-8"?>'
            '<kml xmlns="http://www.opengis.net/kml/2.2" '
                 'xmlns:gx="http://www.google.com/kml/ext/2.2">'
            '<Folder>'
              '<Placemark>'
                '<gx:Track>'
                  '<when>2010-05-28T02:02:09Z</when>'
                  '<when>2010-05-28T02:02:35Z</when>'
                  '<when>2010-05-28T02:02:44Z</when>'
                  '<gx:coord>-122.111111 37.111111 151.333333</gx:coord>'
                  '<gx:coord>-122.222222 37.222222 152.222222</gx:coord>'
                  '<gx:coord>-122.333333 37.333333 153.333333</gx:coord>'
                '</gx:Track>'
              '</Placemark>'
            '</Folder>'
            '</kml>'
        )
        doc = fromstring(test_kml, schema=Schema("kml22gx.xsd"))
        set_max_decimal_places(
            doc, 
            max_decimals={
                'longitude': 3,
                'latitude': 2,
                'altitude': 1,
            }
        )
        
        coords_list = doc.findall(".//{http://www.google.com/kml/ext/2.2}coord")
        #import ipdb; ipdb.set_trace()
        self.assertEquals(
            coords_list[0], 
            '-122.111 37.11 151.3'
        )


from flask import Response, render_template, send_file, jsonify
from geos.kml import *
from geos.print import print_map
from geos import app


def kml_response(kml_map):
    """
    Args:
        kml_map (KMLMap): KMLMap object

    Returns:
        Response: a Flask Response with proper MIME-type.

    """
    return Response(kml_map.get_kml(), mimetype=kml_map.MIME_TYPE)


@app.route('/')
def index():
    return render_template("index.html")


@app.route("/maps.json")
def maps_json():
    """
    Generates a json object which serves as bridge between
    the web interface and the map source collection.

    All attributes relevant for openlayers are converted into
    JSON and served through this route.

    Returns:
        Response: All map sources as JSON object.
    """
    map_sources = {
        id: {
                "id": map_source.id,
                "name": map_source.name,
                "folder": map_source.folder,
                "min_zoom": map_source.min_zoom,
                "max_zoom": map_source.max_zoom,
                "layers": [
                    {
                        "min_zoom": layer.min_zoom,
                        "max_zoom": layer.max_zoom,
                        "tile_url": layer.get_tile_urls,
                    } for layer in map_source.layers
                    ]

            } for id, map_source in app.config["mapsources"].items()
        }
    return jsonify(map_sources)


@app.route('/print/<map_source>/<zoom>/<x>/<y>/<width>/<height>/map.pdf')
def map_to_pdf(map_source, zoom, x, y, width, height):
    """
    Generate a PDF at the given position.

    Args:
        map_source (str): id of the map source to print.
        zoom (int): zoom-level to print
        x (float): Center of the Map in mercator projection (EPSG:4326), x-coordinate
        y (float): Center of the Map in mercator projection (EPSG:4326), y-coordinate
        width (float): width of the pdf in mm
        height (float): height of the pdf in mm

    Returns:

    """
    map_source = app.config["mapsources"][map_source]
    pdf_file = print_map(map_source, x=float(x), y=float(y),
                         zoom=int(zoom), width=float(width), height=float(height), format='pdf')
    return send_file(pdf_file,
                     attachment_filename="map.pdf",
                     as_attachment=True)


@app.route("/kml-master.kml")
def kml_master():
    """KML master document for loading all maps in Google Earth"""
    kml_doc = KMLMaster(app.config["url_formatter"], app.config["mapsources"].values())
    return kml_response(kml_doc)


@app.route("/maps/<map_source>.kml")
def kml_map_root(map_source):
    """KML for a given map"""
    map = app.config["mapsources"][map_source]
    kml_doc = KMLMapRoot(app.config["url_formatter"], map, app.config["LOG_TILES_PER_ROW"])
    return kml_response(kml_doc)


@app.route("/maps/<map_source>/<int:z>/<int:x>/<int:y>.kml")
def kml_region(map_source, z, x, y):
    """KML region fetched by a Google Earth network link. """
    map = app.config["mapsources"][map_source]
    kml_doc = KMLRegion(app.config["url_formatter"], map, app.config["LOG_TILES_PER_ROW"],
                        z, x, y)
    return kml_response(kml_doc)



"""
This module serves for generating Google Earth KML files
which create an overlay of tiled web maps.

This module comes with three major classes:
    * KMLMaster:
        Creates a KML that contains network links to
        multiple KMLMapRoots (-> overview over available maps)
    * KMLMapRoot:
        Root document of a map containing the tiles of the first zoom level.
        Can be used standalone to display one specific map.
    * KMLRegion:
        A region within a KML document containing multiple tiles and f
        our network links to the next zoom level.

The number of tiles per KML region can be specified with the `log_tiles_per_row`
parameter. The number of tiles per region impacts the number of http requests to the server.
Many tiles per region will reduce the amount of KML documents requested while increasing the
number of tiles loaded at once which may be bad for users with a weak internet connection.

Understanding the `log_tiles_per_row` is likely to require some explanation:
    A KMLRegion consists of:
        * always four network links to the next zoom level
        * a certain number of ground overlays (the actual tile images)

    The following constraints apply:
        * A KML region is always a square (nrow = ncol)
        * the number of ground overlays per row is always a power of two.

    `log_tile_per_row` is the log2(tiles per row per region).

        Example: `log_tiles_per_row = 0` -> 2**0 = 1 tile per row -> 1 tile per region
            Network Links           Ground Overlay
                 --- ---               -------
                | 1 | 2 |             |       |
                 --- ---              |   1   |
                | 3 | 4 |             |       |
                 --- ---               -------

        Example: `log_tiles_per_row = 1` -> 2**1 = 2 tilese per row -> four tiles per region
            Network Links           Ground Overlays
                 --- ---               -------
                | 1 | 2 |             | 1 | 2 |
                 --- ---               --- ---
                | 3 | 4 |             | 3 | 4 |
                 --- ---               -------
"""


from pykml_geos.factory import KML_ElementMaker as KML
from geos.geometry import *
from lxml import etree
import math
from abc import ABCMeta
from geos.mapsource import walk_mapsources, F_SEP

DEFAULT_MAX_LOD_PIXELS = -1
DEFAULT_MIN_LOD_PIXELS = 128
MIN_ZOOM_LIMIT = 5  # if minZoom is higher than that, create empty network links.


def kml_element_name(grid_coords, elem_id="KML"):
    """
    Create a unique element name for KML

    Args:
        grid_coords (GridCoordinate):
        elem_id (str):

    >>> kml_element_name(GridCoordinate(zoom=5, x=42, y=60), elem_id="NL")
    'NL_5_42_60'
    """
    return "_".join(str(x) for x in [elem_id, grid_coords.zoom, grid_coords.x, grid_coords.y])


def kml_lat_lon_box(geo_bb):
    """
    Create the north/south/east/west tags for a <LatLonBox> or <LatLonAltBox> Bounding Box

    Args:
        geo_bb (GeographicBB):

    Returns:
        Tuple: (<north>, <south>, <east>, <west>) KMLElements
    """
    return (
        KML.north(geo_bb.max.lat),
        KML.south(geo_bb.min.lat),
        KML.east(geo_bb.max.lon),
        KML.west(geo_bb.min.lon)
    )


def kml_lod(min_lod_pixels=DEFAULT_MIN_LOD_PIXELS, max_lod_pixels=DEFAULT_MAX_LOD_PIXELS):
    """
    Create the KML LevelOfDetail (LOD) Tag.

    In a Region, the <minLodPixels> and <maxLodPixels> elements allow you to specify
    an area of the screen (in square pixels). When your data is projected onto the screen,
    it must occupy an area of the screen that is greater than <minLodPixels> and less
    than <maxLodPixels> in order to be visible. Once the projected size of the Region goes
    outside of these limits, it is no longer visible, and the Region becomes inactive.
    (from https://developers.google.com/kml/documentation/kml_21tutorial)

    Args:
        min_lod_pixels (int):
        max_lod_pixels (int):
    """
    return KML.Lod(
        KML.minLodPixels(min_lod_pixels),
        KML.maxLodPixels(max_lod_pixels))


def kml_region(region_coords, min_lod_pixels=DEFAULT_MIN_LOD_PIXELS,
               max_lod_pixels=DEFAULT_MAX_LOD_PIXELS):
    """
    Create the KML <Region> tag with the appropriate geographic coordinates

    Args:
        region_coords (RegionCoordinate):
        min_lod_pixels (int): see `kml_lod`
        max_lod_pixels (int): see `kml_lod`

    Returns:
        KMLElement: the KML <Region>
    """
    bbox = region_coords.geographic_bounds()
    return KML.Region(
        kml_lod(min_lod_pixels=min_lod_pixels, max_lod_pixels=max_lod_pixels),
        KML.LatLonAltBox(
            *kml_lat_lon_box(bbox)
        )
    )


def kml_network_link(href, name=None, region_coords=None, visible=True):
    """
    Create the KML <NetworkLink> Tag for a certain Region in the RegionGrid.

    Args:
        region_coords (RegionCoordinate):
        href (str): the href attribute of the NetworkLink
        name (str): KML <name>
        visible (bool): If true the network link will appear as 'visible'
            (i.e. checked) in Google Earth.

    Returns:
        KMLElement: the KML <NetworkLink>

    """
    nl = KML.NetworkLink()
    if name is None and region_coords is not None:
        name = kml_element_name(region_coords, "NL")
    if name is not None:
        nl.append(KML.name(name))
    if region_coords is not None:
        min_lod_pixels = DEFAULT_MIN_LOD_PIXELS * (2 ** region_coords.log_tiles_per_row)
        nl.append(kml_region(region_coords, min_lod_pixels=min_lod_pixels))
    if not visible:
        nl.append(KML.visibility(0))

    nl.append(KML.Link(
        KML.href(href), KML.viewRefreshMode("onRegion")))

    return nl


def kml_ground_overlay(tile_coords, tile_url):
    """
    Create a KML <GroundOverlay> for a certain TileCoordinate.

    Args:
        tile_coords (TileCoordinate): TileCoordinate
        tile_url (str): web-url to the actual tile image.

    Returns:
        KMLElement: the KML <GroundOverlay>
    """
    return KML.GroundOverlay(
        KML.name(kml_element_name(tile_coords, "GO")),
        KML.drawOrder(tile_coords.zoom),
        KML.Icon(
            KML.href(tile_url)
        ),
        KML.LatLonBox(
            *kml_lat_lon_box(tile_coords.geographic_bounds())
        ),
    )


def kml_folder(name):
    """
    Create a KML <Folder> tag.

    Args:
        name (str): folder name

    Returns:
        KMLElement: the KML <Folder>
    """
    return KML.Folder(KML.name(name))


class URLFormatter:
    """Create absolute URLs for map resources"""

    def __init__(self, host, port, url_scheme="http"):
        self.host = host
        self.port = port
        self.url_scheme = url_scheme

    def get_abs_url(self, rel_url):
        """
        Create an absolute url from a relative one.

        >>> url_formatter = URLFormatter("example.com", 80)
        >>> url_formatter.get_abs_url("kml_master.kml")
        'http://example.com:80/kml_master.kml'
        """
        rel_url = rel_url.lstrip("/")
        return "{}://{}:{}/{}".format(self.url_scheme, self.host, self.port, rel_url)

    def get_map_root_url(self, mapsource):
        return self.get_abs_url("/maps/{}.kml".format(mapsource.id))

    def get_map_url(self, mapsource, grid_coords):
        """ Get URL to a map region. """
        return self.get_abs_url(
                "/maps/{}/{}/{}/{}.kml".format(mapsource.id, grid_coords.zoom,
                                               grid_coords.x, grid_coords.y))


class KMLMap(metaclass=ABCMeta):
    """Abstract base class representing a KML Document"""
    MIME_TYPE = "application/vnd.google-earth.kml+xml"

    def __init__(self, url_formatter):
        """
        Args:
            url_formatter (URLFormatter): URLFormatter object
        """
        self.url_formatter = url_formatter
        self.kml_doc = KML.Document()
        self.kml_root = KML.kml(self.kml_doc)

    def add_elem(self, kml_elem):
        """Add an element to the KMLDocument"""
        self.kml_doc.append(kml_elem)

    def add_elems(self, kml_elems):
        """
        Add elements from an iterator.

        Args:
            kml_elems (iterable of KMLElements): any iterator containing KML elements.
                Can also be a KMLMap instance
        """
        for kml_elem in kml_elems:
            self.add_elem(kml_elem)

    def get_kml(self):
        """Return the KML Document as formatted kml/xml"""
        return etree.tostring(self.kml_root, pretty_print=True, xml_declaration=True)

    def __iter__(self):
        yield from self.kml_doc.iterchildren()


class KMLMaster(KMLMap):
    """Represents a KML master document that contains NetworkLinks to all maps
    in the mapsource directory."""

    def __init__(self, url_formatter, mapsources):
        """
        Create a KML master document.

        Args:
            mapsources (list of MapSource):
        """
        super().__init__(url_formatter)
        self.map_folders = {
            root: {
                "folders": folders,
                "maps": maps
            } for root, folders, maps in walk_mapsources(mapsources)
        }
        self.add_maps(parent=self.kml_doc)

    def add_maps(self, parent, root_path=""):
        """
        Recursively add maps in a folder hierarchy.

        Args:
            parent (KMLElement): KMLElement to which we want to append child folders or maps respectively
            root_path (str): path of 'parent'
        """
        for mapsource in self.map_folders[root_path]['maps']:
            parent.append(self.get_network_link(mapsource))
        for folder in self.map_folders[root_path]['folders']:
            kml_folder_obj = kml_folder(folder)
            parent.append(kml_folder_obj)
            self.add_maps(parent=kml_folder_obj, root_path=F_SEP.join((root_path, folder)))

    def get_network_link(self, mapsource):
        """Get KML <NetworkLink> for a given mapsource. """
        return kml_network_link(self.url_formatter.get_map_root_url(mapsource),
                                 name=mapsource.name, visible=False)


class KMLMapRoot(KMLMap):
    """Represents the root document of an individual map.
    Can be used as standalone KML to display that map only."""

    def __init__(self, url_formatter, mapsource, log_tiles_per_row):
        """
        Create the root document of an individual map.

        Args:
            mapsource (MapSource):
            log_tiles_per_row (int): see module description. Needs to be in range(0, 5).

        Note:
            The zoom level of the root document is determined as follows:
            The min_zoom level is read from the mapsource. `log_tiles_per_row` defines
            a lower bound for min_zoom. This is because e.g. on zoom level 0 we could not have
            more than one tile per row per region as there is simply only one tile at that zoom
            level.

            However, we can run into severe performance issues, if either min_zoom
            or log_tiles_per_row are too high. At a zoom level of only 8, a root
            document spanning the whole world would already contain (2**8)**2 = 65536 tiles
            which will break map display in Google Earth.

            Therefore MIN_ZOOM_LIMIT is applied as an upper bound. If the determined min_zoom level
            exceeds this limit, empty network links (without GroundOverlay) will be used to adaptively
            load the next zoom level(s).

        """
        super().__init__(url_formatter)
        self.mapsource = mapsource

        assert(log_tiles_per_row in range(0, 5))
        self.log_tiles_per_row = log_tiles_per_row

        # see docstring for explanation
        zoom = min(max(mapsource.min_zoom, log_tiles_per_row), MIN_ZOOM_LIMIT)

        n_tiles = 2 ** zoom                           # tiles per row of the whole document
        tiles_per_row = 2 ** self.log_tiles_per_row   # tiles per row per region
        n_regions = n_tiles//tiles_per_row            # regions per row
        assert n_tiles % tiles_per_row == 0, "regions per row is not an integer."

        if mapsource.bbox is None:
            regions = griditer(0, 0, n_regions)
        else:
            tile_bounds = mapsource.bbox.to_mercator().to_tile(zoom)
            regions = bboxiter(tile_bounds, tiles_per_row)

        self.add_elem(KML.name("{} root".format(mapsource.name)))
        for x, y in regions:
            self.add_elems(KMLRegion(self.url_formatter, self.mapsource,
                                     self.log_tiles_per_row, zoom, x, y))


class KMLRegion(KMLMap):
    """Represents a KML document that is loaded on demand.
    It contains the actual tiles as GroundOverlays and contains NetworkLinks
    to the next LevelOfDetail."""

    def __init__(self, url_formatter, mapsource, log_tiles_per_row, zoom, x, y):
        """
        Create KML document displaying a certain map region.

        Args:
            mapsource (MapSource):
            log_tiles_per_row (int): see module description. Needs to be in range(0, 5).
            zoom (int): zoom level
            x (int): x coordinate identifying the region
            y (int): y coordinate indentifying the region

        Note:
            Will contain four network links to the next zoom level unless zoom = max_zoom.
            Will contain ground overlays with tiles unless zoom < min_zoom
        """
        super().__init__(url_formatter)
        self.mapsource = mapsource
        rc = RegionCoordinate(zoom, x, y, log_tiles_per_row)

        self.add_elem(KML.name(kml_element_name(rc, "DOC")))

        if zoom >= mapsource.min_zoom:
            for tc in rc.get_tiles():
                # add ground overlay for all active layers
                for map_layer in mapsource.layers:
                    if map_layer.min_zoom <= zoom <= map_layer.max_zoom:
                        self.add_ground_overlay(tc, map_layer)

        if zoom < mapsource.max_zoom:
            for rc_child in rc.zoom_in():
                self.add_network_link(rc_child)

    def add_ground_overlay(self, tile_coords, map_layer):
        tile_url = map_layer.get_tile_url(tile_coords.zoom, tile_coords.x, tile_coords.y)
        self.add_elem(kml_ground_overlay(tile_coords, tile_url))

    def add_network_link(self, region_coords):
        href = self.url_formatter.get_map_url(self.mapsource, region_coords)
        self.add_elem(kml_network_link(href, region_coords=region_coords))



"""pyKML Helpers Module

The pykml.helpers module contains 'helper' functions that operate on pyKML 
document objects for accomplishing common tasks.

"""

from pykml.factory import KML_ElementMaker as K
from pykml.factory import GX_ElementMaker as GX

def separate_namespace(qname):
    "Separates the namespace from the element"
    import re
    try:
        namespace, element_name = re.search('^{(.+)}(.+)$', qname).groups()
    except:
        namespace = None
        element_name = qname
    return namespace, element_name

def set_max_decimal_places(doc, max_decimals):
    """Sets the maximum number of decimal places used by KML elements
    
    This method facilitates reducing the file size of a KML document.
    """
    
    def replace_delimited_string_member(
            delimited_str,
            separator,
            index_no,
            decimal_places):
        "Modify the number of decimal places for a delimited string member"
        values = delimited_str.split(separator)
        values[index_no] = str(round(float(values[index_no]), decimal_places))
        return separator.join(values)
    
    if max_decimals.has_key('longitude'):
        data_type = 'longitude'
        index_no = 0 # longitude is in the first position
        # modify <longitude>
        for el in doc.findall(".//{http://www.opengis.net/kml/2.2}longitude"):
            new_val = round(float(el.text), max_decimals[data_type])
            el.getparent().longitude = K.longitude(new_val)
        # modify <coordinates> elements
        for el in doc.findall(".//{http://www.opengis.net/kml/2.2}coordinates"):
            vertex_str_list = []
            for vertex in el.text.strip().split(' '):
                vertex_str_list.append(
                    replace_delimited_string_member(
                        delimited_str=vertex,
                        separator=',',
                        index_no=index_no,
                        decimal_places=max_decimals[data_type]
                    )
                )
            el_new = K.coordinates(' '.join(vertex_str_list).strip())
            el.getparent().replace(el, el_new)
        # modify <gx:coords> elements
        for el in doc.findall(".//{http://www.google.com/kml/ext/2.2}coord"):
            el._setText(
                replace_delimited_string_member(
                    delimited_str=el.text,
                    separator=' ',
                    index_no=index_no,
                    decimal_places=max_decimals[data_type]
                )
            )
    
    if max_decimals.has_key('latitude'):
        data_type = 'latitude'
        index_no = 1 # latitude is in the second position
        # modify <latitude> elements
        for el in doc.findall(".//{http://www.opengis.net/kml/2.2}latitude"):
            new_val = round(float(el.text), max_decimals[data_type])
            el.getparent().latitude = K.latitude(new_val)
        # modify <coordinates> elements
        for el in doc.findall(".//{http://www.opengis.net/kml/2.2}coordinates"):
            vertex_str_list = []
            for vertex in el.text.strip().split(' '):
                vertex_str_list.append(
                    replace_delimited_string_member(
                        delimited_str=vertex,
                        separator=',',
                        index_no=index_no,
                        decimal_places=max_decimals[data_type]
                    )
                )
            el_new = K.coordinates(' '.join(vertex_str_list).strip())
            el.getparent().replace(el, el_new)
        # modify <gx:coords> elements
        for el in doc.findall(".//{http://www.google.com/kml/ext/2.2}coord"):
            el._setText(
                replace_delimited_string_member(
                    delimited_str=el.text,
                    separator=' ',
                    index_no=index_no,
                    decimal_places=max_decimals[data_type]
                )
            )

    if max_decimals.has_key('altitude'):
        data_type = 'altitude'
        index_no = 2 # altitude is in the third position
        # modify <altitude> elements
        for el in doc.findall(".//{http://www.opengis.net/kml/2.2}altitude"):
            new_val = round(float(el.text), max_decimals[data_type])
            el.getparent().altitude = K.altitude(new_val)
        # modify <coordinates> elements
        for el in doc.findall(".//{http://www.opengis.net/kml/2.2}coordinates"):
            vertex_str_list = []
            for vertex in el.text.strip().split(' '):
                vertex_str_list.append(
                    replace_delimited_string_member(
                        delimited_str=vertex,
                        separator=',',
                        index_no=index_no,
                        decimal_places=max_decimals[data_type]
                    )
                )
            el_new = K.coordinates(' '.join(vertex_str_list).strip())
            el.getparent().replace(el, el_new)
        # modify <gx:coords> elements
        for el in doc.findall(".//{http://www.google.com/kml/ext/2.2}coord"):
            el._setText(
                replace_delimited_string_member(
                    delimited_str=el.text,
                    separator=' ',
                    index_no=index_no,
                    decimal_places=max_decimals[data_type]
                )
            )
    
    if max_decimals.has_key('heading'):
        for el in doc.findall(".//{http://www.opengis.net/kml/2.2}heading"):
            new_val = round(float(el.text), max_decimals['heading'])
            el.getparent().heading = K.heading(new_val)
    if max_decimals.has_key('tilt'):
        for el in doc.findall(".//{http://www.opengis.net/kml/2.2}tilt"):
            new_val = round(float(el.text), max_decimals['tilt'])
            el.getparent().tilt = K.tilt(new_val)
    if max_decimals.has_key('range'):
        for el in doc.findall(".//{http://www.opengis.net/kml/2.2}range"):
            new_val = round(float(el.text), max_decimals['range'])
            el.getparent().range = K.range(new_val)



import unittest
from os import path
from lxml import etree
from pykml.parser import Schema
from pykml.factory import KML_ElementMaker as KML
from pykml.factory import ATOM_ElementMaker as ATOM
from pykml.factory import GX_ElementMaker as GX

class KmlFactoryTestCase(unittest.TestCase):
    
    def test_get_factory_object_name(self):
        "Tests obtaining a factory object"
        from pykml.factory import get_factory_object_name
        
        self.assertEqual(
            get_factory_object_name('http://www.opengis.net/kml/2.2'),
            'KML'
        )
        self.assertEqual(
            get_factory_object_name('http://www.w3.org/2005/Atom'),
            'ATOM'
        )
        self.assertEqual(get_factory_object_name(None), 'KML')
    
    def test_trivial_kml_document(self):
        """Tests the creation of a trivial OGC KML document."""
        doc = KML.kml()
        schema = Schema("ogckml22.xsd")
        self.assertTrue(schema.validate(doc))
        self.assertEquals(
            etree.tostring(doc),
            '<kml xmlns:gx="http://www.google.com/kml/ext/2.2" '
                 'xmlns:atom="http://www.w3.org/2005/Atom" '
                 'xmlns="http://www.opengis.net/kml/2.2"/>'
        )
    
    def test_basic_kml_document_2(self):
        """Tests the creation of a basic OGC KML document."""
        doc = KML.kml(
            KML.Document(
                KML.name("KmlFile"),
                KML.Placemark(
                    KML.name("Untitled Placemark"),
                    KML.Point(
                        KML.coordinates("-95.265,38.959,0")
                    )
                )
            )
        )
        self.assertTrue(Schema("kml22gx.xsd").validate(doc))
        self.assertEquals(
            etree.tostring(doc),
            '<kml xmlns:gx="http://www.google.com/kml/ext/2.2" '
                 'xmlns:atom="http://www.w3.org/2005/Atom" '
                 'xmlns="http://www.opengis.net/kml/2.2">'
              '<Document>'
                '<name>KmlFile</name>'
                '<Placemark>'
                  '<name>Untitled Placemark</name>'
                  '<Point>'
                    '<coordinates>-95.265,38.959,0</coordinates>'
                  '</Point>'
                '</Placemark>'
              '</Document>'
            '</kml>'
        )
    
    def test_basic_kml_document(self):
        """Tests the creation of a basic KML with Google Extensions ."""
        doc = KML.kml(
            GX.Tour(
                GX.Playlist(
                    GX.SoundCue(
                        KML.href("http://dev.keyhole.com/codesite/cntowerfacts.mp3")
                    ),
                    GX.Wait(
                        GX.duration(10)
                    ),
                    GX.FlyTo(
                        GX.duration(5),
                        GX.flyToMode("bounce"),
                        KML.LookAt(
                            KML.longitude(-79.387),
                            KML.latitude(43.643),
                            KML.altitude(0),
                            KML.heading(-172.3),
                            KML.tilt(10),
                            KML.range(1200),
                            KML.altitudeMode("relativeToGround"),
                        )
                    )
                )
            )
        )
        self.assertTrue(Schema("kml22gx.xsd").validate(doc))
        self.assertEquals(
            etree.tostring(doc),
            '<kml xmlns:gx="http://www.google.com/kml/ext/2.2" '
                 'xmlns:atom="http://www.w3.org/2005/Atom" '
                 'xmlns="http://www.opengis.net/kml/2.2">'
              '<gx:Tour>'
                '<gx:Playlist>'
                  '<gx:SoundCue>'
                    '<href>http://dev.keyhole.com/codesite/cntowerfacts.mp3</href>'
                  '</gx:SoundCue>'
                  '<gx:Wait>'
                    '<gx:duration>10</gx:duration>'
                  '</gx:Wait>'
                  '<gx:FlyTo>'
                    '<gx:duration>5</gx:duration>'
                    '<gx:flyToMode>bounce</gx:flyToMode>'
                    '<LookAt>'
                      '<longitude>-79.387</longitude>'
                      '<latitude>43.643</latitude>'
                      '<altitude>0</altitude>'
                      '<heading>-172.3</heading>'
                      '<tilt>10</tilt>'
                      '<range>1200</range>'
                      '<altitudeMode>relativeToGround</altitudeMode>'
                    '</LookAt>'
                  '</gx:FlyTo>'
                '</gx:Playlist>'
              '</gx:Tour>'
            '</kml>'
        )
    
    def test_kml_document_with_atom_element(self):
        """Tests the creation of a KML document with an ATOM element."""
        doc = KML.kml(
            KML.Document(
                ATOM.author(
                    ATOM.name("J. K. Rowling")
                ),
                ATOM.link(href="http://www.harrypotter.com"),
                KML.Placemark(
                    KML.name("Hogwarts"),
                    KML.Point(
                        KML.coordinates("1,1")
                    )
                )
            )
        )
        self.assertTrue(Schema("kml22gx.xsd").validate(doc))
        self.assertEquals(
            etree.tostring(doc),
            '<kml xmlns:gx="http://www.google.com/kml/ext/2.2" '
                 'xmlns:atom="http://www.w3.org/2005/Atom" '
                 'xmlns="http://www.opengis.net/kml/2.2">'
              '<Document>'
                '<atom:author>'
                  '<atom:name>J. K. Rowling</atom:name>'
                '</atom:author>'
                '<atom:link href="http://www.harrypotter.com"/>'
                '<Placemark>'
                  '<name>Hogwarts</name>'
                  '<Point>'
                    '<coordinates>1,1</coordinates>'
                  '</Point>'
                '</Placemark>'
              '</Document>'
            '</kml>'
        )

    def test_kml_document_with_cdata_description(self):
        """Tests the creation of a KML document with a CDATA element."""
        from pykml.factory import KML_ElementMaker as KML
        from lxml import etree
        
        doc = KML.description(
                '<h1>CDATA Tags are useful!</h1>'
            )
        self.assertEquals(
            etree.tostring(doc),
              '<description '
                    'xmlns:gx="http://www.google.com/kml/ext/2.2" '
                    'xmlns:atom="http://www.w3.org/2005/Atom" '
                    'xmlns="http://www.opengis.net/kml/2.2">'
                  '&lt;h1&gt;CDATA Tags are useful!&lt;/h1&gt;'
              '</description>'
        )
    
    def test_kml_document_with_cdata_description_2(self):
        """Tests the creation of a KML document with a CDATA element."""
        
        from pykml.factory import KML_ElementMaker as KML
        from pykml.factory import ATOM_ElementMaker as ATOM
        from pykml.factory import GX_ElementMaker as GX
        from lxml import etree
        doc = KML.kml(
          KML.Document(
            KML.Placemark(
              KML.name("CDATA example"),
              KML.description(
                  '<h1>CDATA Tags are useful!</h1>'
                  '<p><font color="red">Text is <i>more readable</i> and '
                  '<b>easier to write</b> when you can avoid using entity '
                  'references.</font></p>'
              ),
              KML.Point(
                KML.coordinates("102.595626,14.996729"),
              ),
            ),
          ),
        )
        self.assertEquals(
            etree.tostring(doc),
            '<kml xmlns:gx="http://www.google.com/kml/ext/2.2" '
                 'xmlns:atom="http://www.w3.org/2005/Atom" '
                 'xmlns="http://www.opengis.net/kml/2.2">'
              '<Document>'
                '<Placemark>'
                  '<name>CDATA example</name>'
                  '<description>'
                    '&lt;h1&gt;CDATA Tags are useful!&lt;/h1&gt;'
                    '&lt;p&gt;&lt;font color="red"&gt;Text is &lt;i&gt;more readable&lt;/i&gt; and '
                    '&lt;b&gt;easier to write&lt;/b&gt; when you can avoid using entity '
                    'references.&lt;/font&gt;&lt;/p&gt;'
                  '</description>'
                  '<Point>'
                    '<coordinates>102.595626,14.996729</coordinates>'
                  '</Point>'
                '</Placemark>'
              '</Document>'
            '</kml>'
        )


class GeneratePythonScriptTestCase(unittest.TestCase):
    
    def test_write_python_script_for_kml_document(self):
        """Tests the creation of a trivial OGC KML document."""
        from pykml.factory import write_python_script_for_kml_document
        
        doc = KML.kml(
            KML.Document(
                ATOM.author(
                    ATOM.name("J. K. Rowling")
                ),
                ATOM.link(href="http://www.harrypotter.com"),
                KML.Placemark(
                    KML.name("Hogwarts"),
                    KML.Point(
                        KML.coordinates("1,1")
                    )
                )
            )
        )
        script = write_python_script_for_kml_document(doc)
        self.assertEquals(
            script,
            'from lxml import etree\n'
            'from pykml.factory import KML_ElementMaker as KML\n'
            'from pykml.factory import ATOM_ElementMaker as ATOM\n'
            'from pykml.factory import GX_ElementMaker as GX\n'
            '\n'
            'doc = KML.kml(\n'
            '  KML.Document(\n'
            '    ATOM.author(\n'
            '      ATOM.name(\'J. K. Rowling\'),\n'
            '    ),\n'
            '    ATOM.link(  href="http://www.harrypotter.com",\n'
            '),\n'
            '    KML.Placemark(\n'
            '      KML.name(\'Hogwarts\'),\n'
            '      KML.Point(\n'
            '        KML.coordinates(\'1,1\'),\n'
            '      ),\n'
            '    ),\n'
            '  ),\n'
            ')\n'
            'print etree.tostring(etree.ElementTree(doc),pretty_print=True)\n'
        )

    def test_write_python_script_for_kml_document_with_cdata(self):
        """Tests the creation of an OGC KML document with a cdata tag"""
        import os
        import tempfile
        from pykml.parser import parse
        from pykml.factory import write_python_script_for_kml_document
        
        test_datafile = path.join(
            path.dirname(__file__),
            'testfiles',
            'google_kml_tutorial/using_the_cdata_element.kml'
        )
        file = 'test/testfiles/'
        schema = Schema('kml22gx.xsd')
        with open(test_datafile) as f:
            doc = parse(f, schema=schema)
        script = write_python_script_for_kml_document(doc)
        self.assertEquals(
            script,
            'from lxml import etree\n'
            'from pykml.factory import KML_ElementMaker as KML\n'
            'from pykml.factory import ATOM_ElementMaker as ATOM\n'
            'from pykml.factory import GX_ElementMaker as GX\n'
            '\n'
            'doc = KML.kml(\n'
            '  KML.Document(\n'
            '    KML.Placemark(\n'
            '      KML.name(\'CDATA example\'),\n'
            '      KML.description(\n'
            '          \'<h1>CDATA Tags are useful!</h1>\'\n'
            '          \'<p><font color="red">Text is <i>more readable</i> and \'\n'
            '          \'<b>easier to write</b> when you can avoid using entity \'\n'
            '          \'references.</font></p>\'\n'
            '      ),\n'
            '      KML.Point(\n'
            '        KML.coordinates(\'102.595626,14.996729\'),\n'
            '      ),\n'
            '    ),\n'
            '  ),\n'
            ')\n'
            'print etree.tostring(etree.ElementTree(doc),pretty_print=True)\n'
        )
        # create a temporary python file
        handle, tfile = tempfile.mkstemp(suffix='.py')
        #print tfile
        with open(tfile, 'w') as f:
            f.write(script)
        
        # execute the temporary python file to create a KML file
        import subprocess
        current_env = os.environ.copy()
        current_env["PYTHONPATH"] = os.path.abspath(os.path.join(os.getcwd(),'..'))
        handle, temp_kml_file = tempfile.mkstemp(suffix='.kml')
        #print temp_kml_file
        with open(temp_kml_file, 'w') as f:
            exit_code = subprocess.call(
                    ["python",tfile],
                    stdout=f,
                    env=current_env
            )
        self.assertEqual(exit_code, 0)
        
        # parse and validate the KML generated by the temporary script
        doc2 = parse(temp_kml_file, schema=schema)
        # test that the root element is as expected
        self.assertEqual(doc2.docinfo.root_name, 'kml')
    
    def test_write_python_script_for_kml_document_with_namespaces(self):
        """Tests the creation of an OGC KML document with several namespaces"""
        import os
        import tempfile
        from pykml.parser import parse
        from pykml.factory import write_python_script_for_kml_document
        
        test_datafile = path.join(
            path.dirname(__file__),
            'testfiles',
            'google_kml_developers_guide/complete_tour_example.kml'
        )
        schema = Schema('kml22gx.xsd')
        with open(test_datafile) as f:
            doc = parse(f, schema=schema)
        script = write_python_script_for_kml_document(doc)
        
        # create a temporary python file
        handle, tfile = tempfile.mkstemp(suffix='.py')
        #print tfile
        with open(tfile, 'w') as f:
            f.write(script)
        
        # execute the temporary python file to create a KML file
        import subprocess
        current_env = os.environ.copy()
        current_env["PYTHONPATH"] = os.path.abspath(os.path.join(os.getcwd(),'..'))
        handle, temp_kml_file = tempfile.mkstemp(suffix='.kml')
        #print temp_kml_file
        with open(temp_kml_file, 'w') as f:
            exit_code = subprocess.call(["python",tfile], stdout=f, env=current_env)
        self.assertEqual(exit_code, 0)
        
        # parse and validate the KML generated by the temporary script
        doc2 = parse(temp_kml_file, schema=schema)
        # test that the root element is as expected
        self.assertEqual(doc2.docinfo.root_name, 'kml')

    def test_write_python_script_for_kml_document_with_comments(self):
        """Tests the creation of an OGC KML document with several comments"""
        import os
        import tempfile
        from pykml.parser import parse
        from pykml.factory import write_python_script_for_kml_document
        
        test_datafile = path.join(
            path.dirname(__file__),
            'testfiles',
            'simple_file_with_comments.kml'
        )
        schema = Schema('kml22gx.xsd')
        with open(test_datafile) as f:
            doc = parse(f, schema=schema)
        script = write_python_script_for_kml_document(doc)
        
        # create a temporary python file
        handle, tfile = tempfile.mkstemp(suffix='.py')
        #print tfile  # Useful for debugging
        with open(tfile, 'w') as f:
            f.write(script)
        
        # execute the temporary python file to create a KML file
        import subprocess
        current_env = os.environ.copy()
        current_env["PYTHONPATH"] = os.path.abspath(os.path.join(os.getcwd(),'..'))
        handle, temp_kml_file = tempfile.mkstemp(suffix='.kml')
        #print temp_kml_file  # Useful for debugging
        with open(temp_kml_file, 'w') as f:
            exit_code = subprocess.call(["python",tfile], stdout=f, env=current_env)
        self.assertEqual(exit_code, 0)
        
        # parse and validate the KML generated by the temporary script
        doc2 = parse(temp_kml_file, schema=schema)
        # test that the root element is as expected
        self.assertEqual(doc2.docinfo.root_name, 'kml')
        
        self.assertEqual(etree.tostring(doc), etree.tostring(doc2))
        
        #import ipdb; ipdb.set_trace()
        pass

if __name__ == '__main__':
    unittest.main()


"""
Default settings that will be loaded in the Flask app.config dict.
"""

# Werkzeug debugger
DEBUG = False

HOST = 'localhost'
PORT = 5000

# between 0 and 5. A higher number will put more tiles in a region, thus reducing
# the number of network links and the server load.
LOG_TILES_PER_ROW = 1




from distutils.core import setup
from setuptools import find_packages

setup(
    name='geos',
    version='0.2.3',
    packages=find_packages(),
    description='Map server to view, measure and print maps in a web browser'
                'and to display maps as an overlay in google earth.',
    author='Gregor Sturm',
    author_email='mail@gregor-sturm.de',
    url='https://github.com/grst/geos',  # use the URL to the github repo
    keywords=['maps', 'google earth', 'overlay', 'map printing'],  # arbitrary keywords
    license='GPLv3',
    install_requires=[
        'flask', 'lxml', 'pillow'
    ],
    extras_require=dict(
        test=['pytest']
    ),
    classifiers=[],
    entry_points={
        'console_scripts': ['geos=geos.scripts.runserver:run_app'],
    },
    include_package_data=True
)



import unittest
from os import path
from pykml.parser import Schema
from pykml.parser import parse

class KmlUtilTestCase(unittest.TestCase):
    
    def test_count_elements(self):
        """Tests the counting of elements in a KML document."""
        from pykml.util import count_elements
        
        test_datafile = path.join(
            path.dirname(__file__),
            'testfiles',
            'google_kml_developers_guide/complete_tour_example.kml'
        )
        with open(test_datafile) as f:
            doc = parse(f, schema=Schema('kml22gx.xsd'))
        summary = count_elements(doc)
        
        self.assertTrue(summary.has_key('http://www.opengis.net/kml/2.2'))
        self.assertEqual(4,
            summary['http://www.opengis.net/kml/2.2']['Placemark']
        )
        self.assertTrue(summary.has_key('http://www.google.com/kml/ext/2.2'))
        self.assertEqual(5,
            summary['http://www.google.com/kml/ext/2.2']['FlyTo']
        )
        self.assertEqual(2,
            summary['http://www.google.com/kml/ext/2.2']['Wait']
        )
